from fastapi import APIRouter

from qualibrate.app.api.routes.branch import branch_router
from qualibrate.app.api.routes.data_file import data_file_router
from qualibrate.app.api.routes.others import other_router
from qualibrate.app.api.routes.project import project_router, projects_router
from qualibrate.app.api.routes.root import root_router
from qualibrate.app.api.routes.snapshot import snapshot_router
from qualibrate.app.api.routes.tag import tag_router

__all__ = [
    "other_router",
    "project_router",
    "projects_router",
    "storage_router",
]

storage_router = APIRouter()

storage_router.include_router(root_router)
storage_router.include_router(branch_router)
storage_router.include_router(data_file_router)
storage_router.include_router(snapshot_router)
storage_router.include_router(tag_router)
